/*
	"display.c" - Sample Code

	Copyright (C) 1998  Sony Computer Entertainment Europe
	All rights Reserved
*/
	// This file contains functions related to the display. This includes
	// loading TIMs, setting up the double buffers, pre-processing ordering
	// tables, special primitives, moving the sky in the pond and showing
	// the end credits.

//#define DEBUG
#define SOURCE_DISPLAY_C

#include <sys/types.h>
#include <libetc.h>
#include <libgte.h>
#include <libgpu.h>
#include <libgs.h>
#include <libmath.h>
#include <libhmd.h>
#include <stdio.h>

#include "3d.h"
#include "main.h"
#include "display.h"
#include "water.h"
#include "music\xmplay.h"

typedef enum { VBLANK_OCCURRED, WAITING_FOR_VBLANK } VBLANK_STATUS;

	// This variable is used to determine when a VSync happens.
	// If it is 1, the VSyncCallBack will drop it to 0, which is
	// a signal for the program to continue. The reason why we
	// do this instead of calling VSync(0) is that VSync(0) will
	// wait until the TV drawing beam is at the TOP of the screen
	// instead of the bottom (when the VBlank actually occurs). This
	// is to fix the GPU bug, but that only manifests itself in interlaced
	// mode. To regain some more CPU time and GPU time, this method is
	// much better.
volatile VBLANK_STATUS waitForVBlank = WAITING_FOR_VBLANK;

void VSyncProcessing(void) {
	if(waitForVBlank == WAITING_FOR_VBLANK) {
		waitForVBlank = VBLANK_OCCURRED;
	}
}

// ================================================================

		// These are special primitives which we will use to change drawing
		// modes and other parameters.

DR_ENV SkyDR_ENV[2];
DR_ENV ScreenBufferDR_ENV[2];
static DR_TWIN ReflectionDR_TWIN;
SPRT skyTiledTexture[2];
static DR_STP
	setSTPon[2],
	setSTPoff;
DR_MODE
	semiTransMinusMode,
	semiTransPlusMode,
	normalMode;

// ================================================================

		// Crunch OT is a function that will remove all NULL GPU commands
		// from an OT. In this demo we were GPU bound, not CPU bound, so it
		// doesn't matter that it is not that efficient.

void CrunchOT(u_long *OT) {

	const u_long
		ADDR_MASK = 0x00ffffff,
		PRIM_MASK = 0xff000000;
	u_long *lastOT = OT;

	OT = (u_long *)((u_long)OT & ADDR_MASK);
	while(1) {
		while(((*OT & ADDR_MASK) != ADDR_MASK) && ((*OT & PRIM_MASK) != 0)) {
			lastOT = OT;
			OT = (u_long *)( (u_long)*OT & ADDR_MASK);
		}
		if((*OT & ADDR_MASK) == ADDR_MASK) {
			break;
		}
		while(((*OT & ADDR_MASK) != ADDR_MASK) && ((*OT & PRIM_MASK) == 0)) {
			OT = (u_long *)( (u_long)*OT & ADDR_MASK);
		}
		*lastOT = (*lastOT & PRIM_MASK) | (u_long)OT;
	}
}

// =====================================================================

	// Call this before drawing up a new screen. Initialises the 3
	// ordering tables.
void initNewDraw(void) {
	GsSetWorkBase(PacketArea[g_ScreenBufferNum]);
	GsClearOt(0,0,&mainOT[g_ScreenBufferNum]);
	GsClearOt(0,0,&groundOT[g_ScreenBufferNum]);
	GsClearOt(0,0,&reflectionOT[g_ScreenBufferNum]);
}

// =====================================================================

	// This function is called when everything that has to be inserted into
	// an ordering table has been inserted, and we are now ready to draw it.

void completeDraw(void) {
	static DISPENV dispEnv =
					  {  { 0, 256, SCREEN_WIDTH, SCREEN_HEIGHT },
						 { 0, 24, 0, SCREEN_HEIGHT-1 },
						 0,0,1,0};

		// Get rid of all NULL packets in the OTs
	CrunchOT((u_long *)reflectionOT[g_ScreenBufferNum].tag);
	CrunchOT((u_long *)groundOT[g_ScreenBufferNum].tag);
	CrunchOT((u_long *)mainOT[g_ScreenBufferNum].tag);

	DrawSync(0);
		// We set waitForVSync to 1 to signal that we are waiting for the
		// VBlank. The callback will reset this to 0.
	waitForVBlank = WAITING_FOR_VBLANK;
	while(waitForVBlank == WAITING_FOR_VBLANK);

		// Change the display environment
	dispEnv.disp.y = 256 - dispEnv.disp.y;
	PutDispEnv(&dispEnv);

		// This will set off a primitive heap which should change drawing
		// environments to the offscreen buffer, set the STP bit on,
		// then draw the sky.
	DrawOTag((u_long *)&SkyDR_ENV[g_ScreenBufferNum]);

		// If we are in the introMode, we need to draw the reflected image
		// before we change drawing environments back to the 'normal' screen
		// buffer.
	if(introModeOn) {
		GsDrawOt(&reflectionOT[g_ScreenBufferNum]); // Draw the reflection in the offscreen buffer
	}

		// This primitive heap will reset the drawing mode and set the STP
		// bit to OFF.
	DrawOTag((u_long *)&normalMode);

		// Set the GPU back to the normal offscreen buffer.
	DrawOTag((u_long *)&ScreenBufferDR_ENV[g_ScreenBufferNum]);

			// In the introMode we now draw the ground, then draw the water
			// on top of it. Otherwise we draw the water (which will just
			// contain the sky), then and put the 'reflection' over it,
			// then draw the ground over that.

	if(introModeOn) {
		GsDrawOt(&groundOT[g_ScreenBufferNum]); // Set GPU in motion.
		if(*firstWaterPrim != 0x00ffffff) {
			DrawOTag(firstWaterPrim);
		}
	} else {
		if(*firstWaterPrim != 0x00ffffff) {
			DrawOTag(firstWaterPrim);
		}
		GsDrawOt(&reflectionOT[g_ScreenBufferNum]); // Draw the reflection in the offscreen buffer
		GsDrawOt(&groundOT[g_ScreenBufferNum]); // Set GPU in motion.
	}

			// Finally, we draw the dinosaur and the innermost tree.
	GsDrawOt(&mainOT[g_ScreenBufferNum]);

	g_ScreenBufferNum ^= 1; // Toggle 'current buffer' number

			// Call the Music Update function
			// We don't call this in VSyncProcessing because we want that
			// to return as soon as possible.
	XM_Update();
}

// =====================================================================

extern DISPENV GsDISPENV;
void initialiseDisplay(void) {
	int i;
	RECT ZeroRect = { 0,0,0,0};
	DRAWENV drawEnv;
	static u_long terminator = 0x00ffffff;
	RECT rect = {0,0,1024,512};
#ifdef DISPLAY_PAL
		SetVideoMode(MODE_PAL);
#else
		SetVideoMode(MODE_NTSC);
#endif

	firstWaterPrim = &terminator;

	GsInitGraph(SCREEN_WIDTH,SCREEN_HEIGHT,GsOFSGPU,1,0);

	ClearImage(&rect,  0,  0,  0); // Clear VRAM

	GsDefDispBuff(0,0,0,SCREEN_HEIGHT);
	GsDISPENV.screen.h = 255;

			// Initialise ordering tables

	mainOT[0].length=mainOT[1].length=MAIN_OT_LENGTH;
	mainOT[0].org=&mainOTTags[0][0];
	mainOT[1].org=&mainOTTags[1][0];

	groundOT[0].length=groundOT[1].length=GROUND_OT_LENGTH;
	groundOT[0].org=&groundOTTags[0][0];
	groundOT[1].org=&groundOTTags[1][0];

	reflectionOT[0].length=reflectionOT[1].length=REFLECTION_OT_LENGTH;
	reflectionOT[0].org=&reflectionOTTags[0][0];
	reflectionOT[1].org=&reflectionOTTags[1][0];

			// Initialise special primitives

	SetDrawStp(&setSTPoff,0);
	termPrim(&setSTPoff);

	SetTexWindow(&ReflectionDR_TWIN,&ZeroRect);
	SetDefDrawEnv(&drawEnv,OFFSCREENBUFFER_X,0,255,SCREEN_HEIGHT-1);
	drawEnv.ofs[0] = OFFSCREENBUFFER_X + 128;
	drawEnv.ofs[1] = OFFSCREENBUFFER_Y + 128;
		// This texture window is used so we can wrap the sky texture
		// which is 64x64 over a larger area. We will reset the texture
		// window to 256x256 after the sky has been drawn.
	drawEnv.tw.x = 896;
	drawEnv.tw.y = 128;
	drawEnv.tw.w = 64;
	drawEnv.tw.h = 64;
		// This tpage is chosen because that's where the sky texture is.
	drawEnv.tpage = getTPage(0,0,896,128);

		// The above DRAWENV structure is used to create two identical
		// special primitives. Two are required so we can adjust one while
		// the GPU is processing the other.

		// Both lists share a common terminator primitive which is never adjusted
		// by the CPU, so there is only one instance of this primitive.

	for(i = 0; i < 2; i++) {
				// Drawing environment
		SetDrawEnv(&SkyDR_ENV[i],&drawEnv);

				// STP bit
		SetDrawStp(&setSTPon[i],1);


				// Sky texture
		setSprt(&skyTiledTexture[i]);
		setRGB0(&skyTiledTexture[i],32,32,32);

		setUV0(&skyTiledTexture[i],0,0);
		setXY0(&skyTiledTexture[i],-128,-128);

		skyTiledTexture[i].w = 254;
		skyTiledTexture[i].h = 255;
		skyTiledTexture[i].clut = getClut(880,434);

				// link them together.

		catPrim(&SkyDR_ENV[i], &setSTPon[i]);
		catPrim(&setSTPon[i],&skyTiledTexture[i]);
		catPrim(&skyTiledTexture[i],&ReflectionDR_TWIN);
	}
	termPrim(&ReflectionDR_TWIN);

	// Now there are two primitive lists:
	// Sky Drawing Env -> STP bit on -> Sky Tiled Texture -> Texture Window reset.

		// ==== Set up the special primitives to change the
		// drawing environment back to the 'normal' screen buffers.

	SetDefDrawEnv(&drawEnv,0,0,SCREEN_WIDTH,SCREEN_HEIGHT);
	drawEnv.ofs[0] = SCREEN_WIDTH/2;
	drawEnv.ofs[1] = SCREEN_HEIGHT/2;
	SetDrawEnv(&ScreenBufferDR_ENV[1],&drawEnv);
	termPrim(&ScreenBufferDR_ENV[1]);

	SetDefDrawEnv(&drawEnv,0,SCREEN_HEIGHT,SCREEN_WIDTH,SCREEN_HEIGHT);
	drawEnv.ofs[0] = SCREEN_WIDTH/2;
	drawEnv.ofs[1] = SCREEN_HEIGHT + SCREEN_HEIGHT/2;
	SetDrawEnv(&ScreenBufferDR_ENV[0],&drawEnv);
	termPrim(&ScreenBufferDR_ENV[0]);

		// initialise special primitives.

	SetDrawMode(&semiTransMinusMode,FALSE,TRUE, GetTPage(2,2,OFFSCREENBUFFER_X,OFFSCREENBUFFER_Y), 0);
	SetDrawMode(&semiTransPlusMode,FALSE,TRUE, GetTPage(2,1,OFFSCREENBUFFER_X,OFFSCREENBUFFER_Y), 0);
	SetDrawMode(&normalMode,FALSE,TRUE, GetTPage(2,0,OFFSCREENBUFFER_X,OFFSCREENBUFFER_Y), 0);

	termPrim(&semiTransMinusMode);
	catPrim(&normalMode,&setSTPoff);
	termPrim(&setSTPoff);

		// Install VSyncCallback handler

	VSyncCallback(VSyncProcessing);

	g_ScreenBufferNum = GsGetActiveBuff();
	SetDispMask(1);
}

// --------------------- TIM Manipulation ----------------------------


// Load TIM will load a TIM into VRAM.
// Parameters:
// 		tim - Points to the very top of the TIM (i.e. including the header)
//		pixelPos - If not NULL, this rect will be used to place the image data.
//		clutPos - If not NULL, this rect will be used to place the CLUT data (if
//				there is any).
//		retImage - If not NULL, copies the image structure to this address
//		retSprite - If not NULL, initialises a basic sprite to this address
void loadTIM(u_long *tim, RECT *pixelPos, RECT *clutPos, GsIMAGE *retImage, GsSPRITE *retSprite) {
	GsIMAGE image;
	RECT p,c;

	GsGetTimInfo(tim + 1,&image);

		// load image
	if(pixelPos) {
		image.px = pixelPos->x;
		image.py = pixelPos->y;
		image.pw = pixelPos->w;
		image.ph = pixelPos->h;
		}
	setRECT(&p,image.px,image.py,image.pw,image.ph);

	LoadImage(&p,image.pixel);

		// load CLUT (if necessary)
	if(image.pmode & 0x8) {
		if(clutPos) {
			image.cx = clutPos->x;
			image.cy = clutPos->y;
			image.cw = clutPos->w;
			image.ch = clutPos->h;
			}
		setRECT(&c,image.cx,image.cy,image.cw,image.ch);
		LoadImage(&c,image.clut);
		}

	if(retImage)
		memcpy(retImage,&image,sizeof(image));

	if(retSprite) {

		retSprite->attribute = //SPRITE_ROT_SCALE_OFF |
							   //	  SPRITE_BRIGHTNESS_OFF |
							   //	  SPRITE_TRANSPARENCY_ON |
							   //	  SPRITE_TRANSPARENCY_MODE(3) |
								  SPRITE_COLOUR_MODE(image.pmode & 3);
		retSprite->x = SCREEN_X_OFFSET;
		retSprite->y = SCREEN_Y_OFFSET;
		switch(image.pmode & 3) {
			case 0:	 retSprite->w = image.pw << 2; break;
			case 1:	 retSprite->w = image.pw << 1; break;
			default: retSprite->w = image.pw;      break;
			}

		retSprite->h = image.ph;
		retSprite->tpage = GetTPage(image.pmode&3,0,image.px,image.py);
		retSprite->u = image.px & 0x3f;  //Mod 64
		retSprite->v = image.py & 0xff;  //Mod 256
		retSprite->cx = image.cx;
		retSprite->cy = image.cy;
		retSprite->r = retSprite->g = retSprite->b = 128;
		retSprite->mx = 0; //retSprite->w/2;
		retSprite->my = 0;//retSprite->h/2;
		retSprite->scalex = retSprite->scaley = ONE;
		retSprite->rotate = 0;
		}

}

// =====================================================================

	// This function will fade the current screen out, show the credits,
	// fade them out, and return. At the end, the screen will be black.

	// The function works by _not_ changing screen buffers, and drawing
	// a single laarge semitrans polygon (100% Background - 100% Foreground)
	// with RGB values of (10,10,10). If this is drawn over the current
	// screen once per frame, it will fade it out. Because we start the
	// semi-trans polygon drawing before the TV beam gets to it, there is
	// no flicker, so double buffering is unnecessary.


void fadeOut(void) {
	int screenBuffer;
	POLY_F4 fadePoly;
	DR_MODE semiTransMinusMode, normalMode;
	DRAWENV drawEnv;
	int fade = 256/4;
	POLY_FT4 title, credits;

	screenBuffer = g_ScreenBufferNum?0:256;

			// =====

	setPolyFT4(&title);
	setUV4(&title, 0, 0,
				   160, 0,
				   0, 65,
				   160, 65);
	title.clut = GetClut(768,498);
	title.tpage = GetTPage(1, 0, 640, 256);
	setXY4(&title, (SCREEN_WIDTH-224)/2,20,
				   SCREEN_WIDTH - (SCREEN_WIDTH-224)/2, 20,
				   (SCREEN_WIDTH-224)/2, 20 + 64,
				   SCREEN_WIDTH - (SCREEN_WIDTH-224)/2, 20 + 64);

	setPolyFT4(&credits);
	setUV4(&credits, 0, 69,
					 224, 69,
					 0, 134,
					 224, 134);
	credits.clut = GetClut(864,423);
	credits.tpage = GetTPage(0, 0, 704, 325);
	setXY4(&credits, (SCREEN_WIDTH-224)/2,150,
					 SCREEN_WIDTH - (SCREEN_WIDTH-224)/2,150,
					 (SCREEN_WIDTH-224)/2,150+65,
					 SCREEN_WIDTH - (SCREEN_WIDTH-224)/2,150+65);

	catPrim(&title,&credits);
	termPrim(&credits);

			// =====

	setPolyF4(&fadePoly);
	setSemiTrans(&fadePoly,1);
	setRGB0(&fadePoly,10,10,10);
	setXY4(&fadePoly,
			0,0,
			SCREEN_WIDTH,0,
			0,400,
			SCREEN_WIDTH, 400);

	SetDrawMode(&semiTransMinusMode,TRUE, FALSE, GetTPage(2,2,0,screenBuffer), 0);
	SetDrawMode(&normalMode,TRUE, FALSE, GetTPage(2,0,0,screenBuffer), 0);

	catPrim(&semiTransMinusMode,&fadePoly);
	termPrim(&fadePoly);

			// =====

	SetDefDrawEnv(&drawEnv,0,screenBuffer, SCREEN_WIDTH, SCREEN_HEIGHT);
	PutDrawEnv(&drawEnv);
	DrawSync(0);
	VSync(0);

		// Fade out main screen

	while(fade > 0) {
		DrawOTag((u_long *)&semiTransMinusMode);
		DrawSync(0);
		VSync(0);
		fade--;
	}

		// Fade in credits
	while(fade <= 128) {
		setRGB0(&credits, fade, fade, fade);
		setRGB0(&title, fade, fade, fade);
		DrawOTag((u_long *)&title);
		fade+=4;
		DrawSync(0);
		VSync(0);
	}

		// Twiddle thumbs for a bit.
	for(fade = 0; fade < 100; fade++) {
		VSync(0);
	}

		// Fade out credits
	fade = 128;
	while(fade > 0) {
		DrawOTag((u_long *)&semiTransMinusMode);
		DrawSync(0);
		VSync(0);
		fade--;
	}
}

// =====================================================================

	// This function will adjust the SPRT's U/V to
	// approximate the movement of the sky. The values passed in
	// are the x and z of the camera position. Strictly speaking,
	// the 'correct' function would also use the camera's height,
	// but this approximation is good enough.

void moveSky(int x, int z) {
	int moveX, moveZ;

		// The value to divide x and z was found by trial and error.
		// Because the texture is tiled 64x64, we mask each value by 63.

	moveX = (u_long)(x / 20) & (64-1);
	moveZ = (u_long)(-z / 20) & (64-1);

	setUV0(&skyTiledTexture[g_ScreenBufferNum],moveX,moveZ);


}