/*
	"water.c" - Sample Code

	Copyright (C) 1998  Sony Computer Entertainment Europe
	All rights Reserved
*/


#define _WATER_

#include <sys/types.h>
#include <libetc.h>
#include <libgte.h>
#include <libgpu.h>
#include <libgs.h>
#include <libhmd.h>
#include <libmath.h>

#include "3d.h"
#include "main.h"
#include "display.h"
#include "water.h"
#include <inline_c.h>

/* Water effect

The pond's surface is made up of a "spider's web" of polygons (12 pie slices
with 4 quads and 1 triangle per slice).

The polygons are textured with using the image in the offscreen buffer.

The polygons do not move, but the U/V coordinates are distorted to give the
illusion of a ripple.

To create the ripple, we use the 'standard' water effect. A rectangular
array of Water Heights is stored, and the horizontal/vertical velocity of
the water flowing _between_ these heights is also stored. If a high Water
Height is adjacent to a low Water Height, then the flow of water will increase
between the two (from high to low), relative the difference between the
two heights.

So the relative difference's between Water Heights affect the flows between
the water, and in turn the flows affect the Water Heights.

To create the illusion, the U/V's of the polygons are distorted using these
flow velocities. By dumping an amount of water on a particular Water Height,
the subsequent iterations create a very realistic water flowing effect.

*/

		// Constants
#define NUM_SECTORS 12
#define QUADS_PER_SECTOR 4


#define PI 3.141592654


#define PREC 16			// num bits of precision for fixed point maths

	// Width & height of water array
#define WATER_WIDTH 12

	// This is a scale factor used to scale the pond's vertices up to
	// fit the environment model.
#define SCALE (125.0 * 7.57 / 128.0)

// Globals

static const u_long terminator = 0x00ffffff;

		// The array of water heights.
static int Water[WATER_WIDTH][WATER_WIDTH];

		// The vertical and horizontal velocities between water heights.
static int VelX[WATER_WIDTH*(WATER_WIDTH-1)];
static int VelY[(WATER_WIDTH-1)*WATER_WIDTH];

		// Primitive heaps. There are 3 copies. 2 are for the CPU/GPU (double
		// buffered) and the last one is a reference for the original U/Vs.
		// Nothing else is used in the last copy. This is very memory
		// inefficient, but we've got a lot to play around with!

		// The primitives are divided up into pie slices (sectors). There
		// are many quads and exactly one triangle per sector.
static POLY_FT4 waterPrimHeap4[3][NUM_SECTORS][QUADS_PER_SECTOR];
static POLY_FT3 waterPrimHeap3[3][NUM_SECTORS];

		// These form an indirect reference array. Since a vertex point doesn't
		// easily map to a water velocity, it must instead map to the nearest
		// velocity. The arrays below hold indexes showing which vertex maps
		// to which water velocity. These arrays are precalculated at the
		// start of the program.
static u_long waterPrimHeap4_Ref[NUM_SECTORS][QUADS_PER_SECTOR][4][2];	// Num angles, num polys, num verts, X/Y
static u_long waterPrimHeap3_Ref[NUM_SECTORS][3][2];  // Num angles, num verts, X/Y

	// The vertices to be transformed. These are precalculated at the beginning.
	// The transformed vertices are stored on the scratchpad.
SVECTOR waterVert[NUM_SECTORS][QUADS_PER_SECTOR+2];

// ===============================================================
// ====== ADD LUMP ===============================================
// ===============================================================
	// This function will add a 'lump' of water onto the water heights. This
	// will cause a ripple to appear in the water. Effort has been made
	// to keep the total amount of water equal at all times. If we added
	// water adinfinitum, the water heights would potentially overflow
	// after a long time.
void AddLump(int size) {
	int xx,yy, redux;

	size <<= PREC;
	redux = size / (WATER_WIDTH * WATER_WIDTH);
	for(yy = 0; yy < WATER_WIDTH; yy++) {
		for(xx = 0; xx < WATER_WIDTH; xx++) {
			Water[yy][xx] -= redux;
		}
	}
	Water[WATER_WIDTH/2][WATER_WIDTH/2] += size;
}


// =====================================================================
// ==== RENDER WATER ===================================================
// =====================================================================
	// Assumes the GTE has been set up. This calculates the flows and
	// feedbacks, then creates the primitive list for the pond.

	// The texture for this is the offscreen buffer. In Intro Mode, this
	// contains any models which would be reflected (the pteradactyl, the
	// dino, the inner tree and the sky). In Main Mode, it contains only
	// the sky.

void RenderWater(void) {
	int i,j,x,y;
	SVECTOR *inVert;
	u_long flg;
	u_long *outVertValidFlagBase, *outVert;

	POLY_FT4 *primHeap4, *primHeap4Original;
	u_long *vertexReference4;

	POLY_FT3 *primHeap3, *primHeap3Original;
	u_long *vertexReference3;

	u_long *lastPrim,*currPrim;
	u_long RGBcode;

	int *Vel,*VelX_p, *VelY_p, *Water_p, lastVelX, *lastVelY_p;

		// Since we only ripple during the intro mode, we only need to
		// calculate the water changes in intro mode.
	if(introModeOn) {

#define DAMPER (2 * (1 << PREC))

				// Calculate the change in the velocity of the flow.
		Vel = &VelX[0];
		for(y = 0; y < WATER_WIDTH; y++) {
			for(x = 0; x  < WATER_WIDTH - 1; x++) {
				*Vel += (Water[y][x+1] - Water[y][x])/2;
				if(*Vel > Water[y][x+1]) {
					*Vel = Water[y][x+1];
				} else if(*Vel < -Water[y][x]) {
					*Vel = -Water[y][x];
				}
				if(*Vel > -DAMPER && *Vel < DAMPER) {
					*Vel = 0;
				}
				Vel++;
			}
		}

		Vel = &VelY[0];
		for(y = 0; y < WATER_WIDTH - 1; y++) {
			for(x = 0; x  < WATER_WIDTH; x++) {
				*Vel += (Water[y+1][x] - Water[y][x])/2;
				if(*Vel > Water[y+1][x]) {
					*Vel = Water[y+1][x];
				} else if(*Vel < -Water[y][x]) {
					*Vel = -Water[y][x];
				}
				if(*Vel > -DAMPER && *Vel < DAMPER) {
					*Vel = 0;
				}
				Vel++;
			}
		}

				// Use the flow velocities to affect the Water Heights.
				// This is hideous looking code, but it means:
				//		We only read/write each Water height once
				//		We only read and divide each Velocity once.

		for(i = 0; i < WATER_WIDTH ; i++) {		// Clear D-Cache
			(getScratchAddr(0))[i] = 0;
		}

		VelX_p = &VelX[0];
		VelY_p = &VelY[0];
		Water_p = &Water[0][0];

		for(i = 0; i < WATER_WIDTH-1 ; i++) {
			int VelY_v;
			lastVelX = 0;
			lastVelY_p = (int *)getScratchAddr(0);
			for(j = 0; j < WATER_WIDTH - 1; j++) {
				int VelX_v = *(VelX_p++)/2;
					VelY_v = *(VelY_p++)/2;
				*(Water_p++) += VelX_v - lastVelX + VelY_v - *lastVelY_p;
				*(lastVelY_p++) = VelY_v;
				lastVelX = VelX_v;
			}
			VelY_v = *(VelY_p++)/2;
			*(Water_p++) += VelY_v - *lastVelY_p - lastVelX;
			*lastVelY_p = VelY_v;
		}
		lastVelX = 0;
		lastVelY_p = (int *)getScratchAddr(0);
		for(j = 0; j < WATER_WIDTH - 1; j++) {
			int VelX_v = *(VelX_p++)/2;
			*(Water_p++) += VelX_v - lastVelX - *(lastVelY_p++);
			lastVelX = VelX_v;
		}
		*Water_p -= *lastVelY_p + lastVelX;

	}	// end if(introModeOn)

	// ======================================================
		// RENDER
		// This section renders the pond vertices.
		// The primitive heaps already contain the correct clut, tpage
		// and RGB values. The primitive linked list must be built every frame
		// (because some polygons may be clipped), and the primitive code
		// must be adjusted depending on the introMode because it should
		// be semi-trans in Intro Mode.

		// Transform all the verts and throw them into the D-Cache
		// A bitmask which stores whether there were any errors with the
		// transform is also stored in the D-Cache.
#define VALIDBASE ((u_long *)&getScratchAddr(0)[(NUM_SECTORS+1) * (QUADS_PER_SECTOR+2)])

	inVert = &waterVert[0][0];
		// Place a bit field here to indicate whether the vertex is valid
	outVertValidFlagBase = VALIDBASE;
	outVert = getScratchAddr(0);	// Place the transformed verts here.
		// Transform all vertices into screen coords
	for(j = 0; j < NUM_SECTORS ; j++) {
		u_long orMask = 0x7; // = Binary 111
		u_long valid = 0;
		for(i = 0; i < (QUADS_PER_SECTOR+2)/3; i++) {
			gte_ldv3c(inVert);
			gte_rtpt();								// Matrix multiply
			gte_stflg(&flg);						// Store the error flag
			gte_stsxy3c(outVert);	// Save the vertices

			if(flg < 0) {
				valid |= orMask;
			}
			inVert += 3;
			outVert += 3;
			orMask <<= 3;
		}
		*(outVertValidFlagBase++) = valid;
	}
		// Recopy these for 'wrap'
	for(j = 0; j < QUADS_PER_SECTOR+2; j++) {
		*(outVert++) = getScratchAddr(0)[j];
	}
	*outVertValidFlagBase = *VALIDBASE;


			// Now we make a primitive list, excluding all the primitives that
			// have vertices which the GTE didn't like.

	outVertValidFlagBase = VALIDBASE;
	outVert = getScratchAddr(0);	// Place the transformed verts here.

		// Here's a handy optimisation when building primitive lists. The
		// address part of the primitive is just the lower 24 bits of the
		// main memory address. This means that if your code is compiled
		// to segment 0x80000000, you have to keep 'and'ing it with 0x00ffffff
		// to get the address to insert it into the primitive. However, since
		// segment 0x80000000 is identical to segement 0x0, you don't
		// have to do the 'and' if you've either compiled to segment 0 or
		// if you and the addresses when you initialise (as we do below).

		// This means not using catPrim, which always does the 'and'.

	primHeap4 = (POLY_FT4 *)((u_long)&waterPrimHeap4[g_ScreenBufferNum][0][0] & 0xffffff);
	primHeap4Original = &waterPrimHeap4[2][0][0];
	vertexReference4 = &waterPrimHeap4_Ref[0][0][0][0];

	primHeap3 = (POLY_FT3 *)((u_long)&waterPrimHeap3[g_ScreenBufferNum][0] & 0xffffff);
	primHeap3Original = &waterPrimHeap3[2][0];
	vertexReference3 = &waterPrimHeap3_Ref[0][0][0];

	lastPrim = (u_long *)((u_long)&terminator & 0xffffff);

	for(i = 0; i < NUM_SECTORS; i++) {
		RGBcode = (0x2c << 24) | (0x808080);
		if(introModeOn) {
			RGBcode |= 0x02000000;	// Set semitrans on
		}
		for(j = 0; j < QUADS_PER_SECTOR; j++) {
			if(((*outVertValidFlagBase >> j) & 0x3) ||
			   ((*(outVertValidFlagBase+1) >> j) & 0x3)) {
				// Do nothing if any one of the verts is invalid.
				vertexReference4 += 2*4;
			} else {
				currPrim = (u_long *)primHeap4;
				currPrim[0] = (9 << 24) | (u_long)lastPrim;
				currPrim[1] = RGBcode;

				lastPrim = currPrim;
					// This also fills in the Y.
				*(u_long *)&primHeap4->x0 = outVert[0];
				*(u_long *)&primHeap4->x2 = outVert[1];
				*(u_long *)&primHeap4->x1 = outVert[QUADS_PER_SECTOR + 2];
				*(u_long *)&primHeap4->x3 = outVert[QUADS_PER_SECTOR + 2 + 1];
					// Must be in this order for it to work!
				primHeap4->u0 = primHeap4Original->u0 + *(VelX + *(vertexReference4++)) / 2 / (1<< PREC);
				primHeap4->v0 = primHeap4Original->v0 + *(VelY + *(vertexReference4++)) / 2 / (1<< PREC);
				primHeap4->u1 = primHeap4Original->u1 + *(VelX + *(vertexReference4++)) / 2 / (1<< PREC);
				primHeap4->v1 = primHeap4Original->v1 + *(VelY + *(vertexReference4++)) / 2 / (1<< PREC);
				primHeap4->u2 = primHeap4Original->u2 + *(VelX + *(vertexReference4++)) / 2 / (1<< PREC);
				primHeap4->v2 = primHeap4Original->v2 + *(VelY + *(vertexReference4++)) / 2 / (1<< PREC);
				primHeap4->u3 = primHeap4Original->u3 + *(VelX + *(vertexReference4++)) / 2 / (1<< PREC);
				primHeap4->v3 = primHeap4Original->v3 + *(VelY + *(vertexReference4++)) / 2 / (1<< PREC);
			}
			primHeap4Original++;
			primHeap4++;

			outVert++;
		}

				// Triangle
		RGBcode = (0x24 << 24) | (0x808080);
		if(introModeOn) {
			RGBcode |= 0x02000000;	// Set semitrans on
		}

		if(((*outVertValidFlagBase >> QUADS_PER_SECTOR) & 0x3) ||
		   ((*(outVertValidFlagBase+1) >> QUADS_PER_SECTOR) & 0x2)) {
			// Do nothing if any one of the verts is invalid.
			vertexReference3 += 2*3;
		} else {
			currPrim = (u_long *)primHeap3;
			currPrim[0] = (7 << 24) | (u_long)lastPrim;
			currPrim[1] = RGBcode;

			lastPrim = currPrim;
				// This also fills in the Y.
			*(u_long *)&primHeap3->x0 = *(outVert);
			*(u_long *)&primHeap3->x1 = *(outVert + (QUADS_PER_SECTOR + 2));
			*(u_long *)&primHeap3->x2 = *(outVert + 1);

					// These must be in this order for the waterReference reference to work.
			primHeap3->u0 = primHeap3Original->u0 + *(VelX + *(vertexReference3++)) / 2 / (1<< PREC);
			primHeap3->v0 = primHeap3Original->v0 + *(VelY + *(vertexReference3++)) / 2 / (1<< PREC);
			primHeap3->u1 = primHeap3Original->u1 + *(VelX + *(vertexReference3++)) / 2 / (1<< PREC);
			primHeap3->v1 = primHeap3Original->v1 + *(VelY + *(vertexReference3++)) / 2 / (1<< PREC);
			primHeap3->u2 = primHeap3Original->u2 + *(VelX + *(vertexReference3++)) / 2 / (1<< PREC);
			primHeap3->v2 = primHeap3Original->v2 + *(VelY + *(vertexReference3++)) / 2 / (1<< PREC);

		}
		primHeap3++;
		primHeap3Original++;

		outVertValidFlagBase++;
		outVert += 2;
	}

	firstWaterPrim = lastPrim;
}

// =====================================================================
// ==== INITIALISE WATER ===============================================
// =====================================================================

static void findNearestPoint(int xPoint, int yPoint, int *retX, int *retY) {

	int	minDistance = 1000000;
	int x,y;
	int potentialX,potentialY,deltaX,deltaY,distance;

	for(x = 0; x < WATER_WIDTH-1; x++) {
		for(y = 0; y < WATER_WIDTH-1; y++) {
			potentialX = x * 256 / (WATER_WIDTH-1);
			potentialY = y * 256 / (WATER_WIDTH-1);

			deltaX = potentialX - xPoint;
			if(deltaX < 0) deltaX = -deltaX;

			deltaY = potentialY - yPoint;
			if(deltaY < 0) deltaY = -deltaY;

			distance = deltaX + deltaY;
			if(distance < minDistance) {
				minDistance = distance;
				*retX = x;
				*retY = y;
			}
		}
	}
}


void initWater(void) {
	int x,y,z, angle,i,j,k;
	int *Vel;
	int innerLimit, outerLimit;
	int sector;
	int xIndex,yIndex;

	POLY_FT4 *prim4 = &waterPrimHeap4[0][0][0];
	POLY_FT3 *prim3 = &waterPrimHeap3[0][0];

	for(y = 0; y < WATER_WIDTH; y++) {
		for(x = 0; x < WATER_WIDTH; x++) {
			Water[y][x] = 10 << PREC;
		}
	}

	for(Vel = &VelX[0], y = 0; y < WATER_WIDTH; y++) {
		for(x = 0; x < WATER_WIDTH-1; x++) {
			*(Vel++) = 0;
		}
	}

	for(Vel = &VelY[0], y = 0; y < WATER_WIDTH-1; y++) {
		for(x = 0; x < WATER_WIDTH; x++) {
			*(Vel++) = 0;
		}
	}

	for(angle = 0,sector = 0; sector < NUM_SECTORS ; angle += 4096 / NUM_SECTORS,sector++) {
		outerLimit = 128;
		innerLimit = outerLimit - 128/(QUADS_PER_SECTOR+1);

		for(z = 0; z < QUADS_PER_SECTOR; z++) {
			setPolyFT4(prim4);
			setSemiTrans(prim4,1);
			setRGB0(prim4,0x80,0x80,0x80);
			prim4->clut = 0;
			prim4->tpage = getTPage(2,	// 16 bit direct
								  1,	// Transparency : 1 = 100% B + 100% F
								  OFFSCREENBUFFER_X,	// X position
								  OFFSCREENBUFFER_Y);	// Y position

			prim4->y0 = -outerLimit * rcos(angle) / 4096;
			prim4->y1 = -outerLimit * rcos(angle+4096/NUM_SECTORS) / 4096;
			prim4->y2 = -innerLimit * rcos(angle) / 4096;
			prim4->y3 = -innerLimit * rcos(angle+4096/NUM_SECTORS) / 4096;

			prim4->x0 = outerLimit * rsin(angle) / 4096;
			prim4->x1 = outerLimit * rsin(angle+4096/NUM_SECTORS) / 4096;
			prim4->x2 = innerLimit * rsin(angle) / 4096;
			prim4->x3 = innerLimit * rsin(angle+4096/NUM_SECTORS) / 4096;

			prim4->u0 = ((prim4->x0 + 128) > 255)?255:(prim4->x0 + 128);
			prim4->u1 = ((prim4->x1 + 128) > 255)?255:(prim4->x1 + 128);
			prim4->u2 = ((prim4->x2 + 128) > 255)?255:(prim4->x2 + 128);
			prim4->u3 = ((prim4->x3 + 128) > 255)?255:(prim4->x3 + 128);

			prim4->v0 = ((prim4->y0 + 128) > 255)?255:(prim4->y0 + 128);
			prim4->v1 = ((prim4->y1 + 128) > 255)?255:(prim4->y1 + 128);
			prim4->v2 = ((prim4->y2 + 128) > 255)?255:(prim4->y2 + 128);
			prim4->v3 = ((prim4->y3 + 128) > 255)?255:(prim4->y3 + 128);

			waterVert[sector][z].vx = -prim4->x0 * SCALE;
			waterVert[sector][z].vy = WATER_DIST_FROM_0;
			waterVert[sector][z].vz = prim4->y0 * SCALE;

			outerLimit = innerLimit;
			innerLimit -= 128 / (QUADS_PER_SECTOR + 1);
			prim4++;
		}
		setPolyFT3(prim3);
		setSemiTrans(prim3,1);
		setRGB0(prim3,0x80,0x80,0x80);
		prim3->clut = 0;
		prim3->tpage = getTPage(2,	// 16 bit direct
							  1,	// Transparency - 1 = 100% B + 100% F
							  OFFSCREENBUFFER_X,	// X position
							  OFFSCREENBUFFER_Y);	// Y position

		prim3->y0 = -outerLimit * rcos(angle) / 4096;
		prim3->y1 = -outerLimit * rcos(angle+4096/NUM_SECTORS) / 4096;
		prim3->y2 = -0;

		prim3->x0 = outerLimit * rsin(angle) / 4096;
		prim3->x1 = outerLimit * rsin(angle+4096/NUM_SECTORS) / 4096;
		prim3->x2 = 0;

		prim3->u0 = ((prim3->x0 + 128) > 255)?255:(prim3->x0 + 128);
		prim3->u1 = ((prim3->x1 + 128) > 255)?255:(prim3->x1 + 128);
		prim3->u2 = ((prim3->x2 + 128) > 255)?255:(prim3->x2 + 128);

		prim3->v0 = ((prim3->y0 + 128) > 255)?255:(prim3->y0 + 128);
		prim3->v1 = ((prim3->y1 + 128) > 255)?255:(prim3->y1 + 128);
		prim3->v2 = ((prim3->y2 + 128) > 255)?255:(prim3->y2 + 128);

		waterVert[sector][QUADS_PER_SECTOR].vx = -prim3->x0 * SCALE;
		waterVert[sector][QUADS_PER_SECTOR].vy = WATER_DIST_FROM_0;
		waterVert[sector][QUADS_PER_SECTOR].vz = prim3->y0 * SCALE;

		waterVert[sector][QUADS_PER_SECTOR+1].vx = 0;
		waterVert[sector][QUADS_PER_SECTOR+1].vy = WATER_DIST_FROM_0;
		waterVert[sector][QUADS_PER_SECTOR+1].vz = 0;

		prim3++;
	}

			// Duplicate the primitive heap

	for(i = 1; i < 3; i++) {
		prim4 = &waterPrimHeap4[i][0][0];
		prim3 = &waterPrimHeap3[i][0];
		for(sector = 0; sector < NUM_SECTORS; sector++) {
			for(z = 0; z < QUADS_PER_SECTOR; z++) {
				*(prim4++) = waterPrimHeap4[0][sector][z];
			}
			*(prim3++) = waterPrimHeap3[0][sector];
		}
	}

	//    Mapper
	// For each primitive, designates which UV point maps to which Velocity
	// height.

	for(i = 0; i < NUM_SECTORS; i++) {
		for(j = 0; j < QUADS_PER_SECTOR; j++) {
			for(k = 0; k < 4; k++) {
				switch(k) {
					case 0:
						findNearestPoint(waterPrimHeap4[0][i][j].u0,waterPrimHeap4[0][i][j].v0,&xIndex,&yIndex);
						break;
					case 1:
						findNearestPoint(waterPrimHeap4[0][i][j].u1,waterPrimHeap4[0][i][j].v1,&xIndex,&yIndex);
						break;
					case 2:
						findNearestPoint(waterPrimHeap4[0][i][j].u2,waterPrimHeap4[0][i][j].v2,&xIndex,&yIndex);
						break;
					case 3:
						findNearestPoint(waterPrimHeap4[0][i][j].u3,waterPrimHeap4[0][i][j].v3,&xIndex,&yIndex);
						break;
				}
				waterPrimHeap4_Ref[i][j][k][0] = xIndex + (WATER_WIDTH-1) * yIndex;
				waterPrimHeap4_Ref[i][j][k][1] = xIndex + WATER_WIDTH * yIndex;
			}
		}
				// Now do the triangle

		for(k = 0; k < 3; k++) {
			switch(k) {
				case 0:
					findNearestPoint(waterPrimHeap3[0][i].u0,waterPrimHeap3[0][i].v0,&xIndex,&yIndex);
					break;
				case 1:
					findNearestPoint(waterPrimHeap3[0][i].u1,waterPrimHeap3[0][i].v1,&xIndex,&yIndex);
					break;
				case 2:
					findNearestPoint(waterPrimHeap3[0][i].u2,waterPrimHeap3[0][i].v2,&xIndex,&yIndex);
					break;
			}
			waterPrimHeap3_Ref[i][k][0] = xIndex + (WATER_WIDTH-1) * yIndex;
			waterPrimHeap3_Ref[i][k][1] = xIndex + WATER_WIDTH * yIndex;
		}
	}
}